////////////////////////////////////////////////////////////////////////
//
//      Fichier: Barre_impl.cpp
//       Auteur: Stephane Drouin
//     Creation: 20/03/01 12:07:54
// Modification: 
//  Description: See Barre_impl.h
//
////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <OB/CORBA.h>
#include <Barre_impl.h>

////////////////////////////////////////////////////////////////////////
// Destructeur d'une barre
////////////////////////////////////////////////////////////////////////
Barre_impl::~Barre_impl() { 
	cout << "Destruction d'une barre..." << endl;	// Pour signaler la vie!
	if (grille) delete [] grille;
}

////////////////////////////////////////////////////////////////////////
// Le client doit detruire le servant apres utilisation
////////////////////////////////////////////////////////////////////////
void Barre_impl::Detruire() {
	// Code tire de la doc de ORBacus, page 91
	PortableServer::POA_var poa = _default_POA();
	PortableServer::ObjectId_var id = poa->servant_to_id(this);
	poa->deactivate_object(id);

	// When the reference count reaches a value of of the zero,
	// the _remove_ref code will issue the call delete this.
	// http://phoenix.liu.edu/~mdevi/corba/RefCountServantBase.htm
	_remove_ref();
}

////////////////////////////////////////////////////////////////////////
// Initialisation et validation des parametres
////////////////////////////////////////////////////////////////////////
void Barre_impl::Initialisation(const ParametresSimulation& par) {
	cout << "Initialisation d'une barre" << endl;	// Pour signaler la vie!

	// Validation des parametres de simulation touchant ce canal
	p = par;
	if (p.chin<0) throw MauvaisParametres(p,"Chaleur generee negative");
	if (p.diffusion<0) throw MauvaisParametres(p,"Diffusion de chaleur negative");
	if (p.chspecb<=0) throw MauvaisParametres(p,"chspecb doit etre positif");
	if (p.densb<=0) throw MauvaisParametres(p,"Densite barre doit etre positive");
	if (p.longb<=0) throw MauvaisParametres(p,"Longueur barre doit etre positive");
	if (p.largb<=0) throw MauvaisParametres(p,"Largeur barre doit etre positive");
	if (p.ngrilles<3) throw MauvaisParametres(p,"Grille: 3 pts sont necessaires");

	// Initialisation des valeurs de l'objet
	if (grille) delete [] grille;
	grille = new TempType[p.ngrilles*p.ngrilles];
	int i;
	for (i=0;i<p.ngrilles*p.ngrilles;++i) grille[i] = p.tempin;
	for (i=0;i<4;++i) gradient[i] = 0;
	maxtemp = p.tempin;
}

////////////////////////////////////////////////////////////////////////
// Obtenir la temperature maximale de la barre
////////////////////////////////////////////////////////////////////////
TempType Barre_impl::temp_max() { return maxtemp; }

////////////////////////////////////////////////////////////////////////
// Obtenir les gradients de temperature de la barre
///////////////////////////////////////////////////////////////////////
Barre::TempArray_slice* Barre_impl::grad_temp() { return gradient; }

////////////////////////////////////////////////////////////////////////
// Calcul des nouvelles temperatures moyenne de la barre
////////////////////////////////////////////////////////////////////////
void Barre_impl::Iteration(TimeType pas, const Barre::TempArray temp_cal) {
	//////////////
	// Calcul de toutes les nouvelles temperatures
	//////////////
	TempType deltaT;				// Difference de temperature
	TempType d2T;					// Laplacien en un point
	TempType *nouvellegrille;       // Temperatures calculees a t+1
	nouvellegrille = new TempType[p.ngrilles*p.ngrilles];

	// Largeur des points sur la grille (au carre)
	const PhysicsType dx2 = (p.largb*p.largb)/(p.ngrilles*p.ngrilles);

	IndiceType i,j;
	maxtemp = 0;
	for (i=0;i<p.ngrilles;++i) for (j=0;j<p.ngrilles;++j) {	// Calcul sur la grille
		d2T = -4*grille[i*p.ngrilles+j];	// Formule des diffrences finies

		// Cas particulier si sur le bord de la grille: face 0,1,2,3?
		d2T += ((i>0)            ? grille[(i-1)*p.ngrilles+j] : temp_cal[1]); //0
		d2T += ((i<p.ngrilles-1) ? grille[(i+1)*p.ngrilles+j] : temp_cal[3]); //2
		d2T += ((j>0)            ? grille[i*p.ngrilles+j-1]   : temp_cal[0]); //1
		d2T += ((j<p.ngrilles-1) ? grille[i*p.ngrilles+j+1]   : temp_cal[2]); //3

		// Division du total par la largeur au carre
		d2T /= dx2;

		// Calcul de la nouvelle temperature
		deltaT = (p.diffusion*d2T+p.chin)*pas/(p.chspecb*p.densb);
		nouvellegrille[i*p.ngrilles+j] = grille[i*p.ngrilles+j]+deltaT;

		if (nouvellegrille[i*p.ngrilles+j]>maxtemp) {	// Cherche temp maximale
			maxtemp=nouvellegrille[i*p.ngrilles+j];
		}

		// L'instabilitee numerique est detectee si temperature negative:
		// physiquement impossible!
		if (nouvellegrille[i*p.ngrilles+j] < 0) {
			throw ErreurCalcul("Instabilite numerique detectee");
		}
	}
	delete [] grille;			// Nouvelles temperatures completes
	grille = nouvellegrille;	// Changement final de grille

	//////////////
	// Calcul du gradient
	//////////////
	const PhysicsType dx = p.largb/p.ngrilles;	// Largeur d'un point
	for (i=0;i<4;++i) gradient[i] = 0;
	for (i=1;i<p.ngrilles-1;++i) {	// Faces verticale: i
		gradient[0] += grille[i*p.ngrilles+1]-grille[i*p.ngrilles];	// Face 1
		gradient[2] += grille[i*p.ngrilles+p.ngrilles-2]
						-grille[i*p.ngrilles+p.ngrilles-1];			// Face 3
	}
	for (j=1;j<p.ngrilles-1;++j) {	// Faces horizontales: j
	    gradient[1] += grille[p.ngrilles+j]-grille[j];				// Face 2
		gradient[3] += grille[(p.ngrilles-2)*p.ngrilles+j]
						-grille[(p.ngrilles-1)*p.ngrilles+j];		// Face 4
	}
	for (i=0;i<4;++i) gradient[i] /= ((p.ngrilles-2)*dx);	// Moyenne gradients
}
