/////////////////////////////////////////////////////////////////////////////
//
//      Fichier: Reacteur_impl.cpp
//       Auteur: Stephane Drouin
//     Creation: 21/02/01 21:56:57
// Modification: 28/02/01 18:44:25
//               Correction de calop(...) : cas k=2. Remplace la valeur
//               12 (nbarres=3) par p.nbarres*p.nbarres+p.nbarres.
// Modification: 21/03/01
//				 Ajout des exception pertinentes.
// Modification: 04/04/01
//               Ajout des fabriques pour creer les barres et canaux
//  Description: See Reacteur_impl.hpp
//
/////////////////////////////////////////////////////////////////////////////

#include "Reacteur_impl.h"
#include <OB/CosNaming.h>
#include <iostream.h>

/////////////////////////////////////////////////////////////////////////////
// Constructeur d'un Reacteur
/////////////////////////////////////////////////////////////////////////////
Reacteur_impl::Reacteur_impl(CORBA::ORB_ptr _orb) : barres(0), canaux(0){
	orb = _orb;		// Un pointeur sur l'ORB est conserve dans le simulateur
}

/////////////////////////////////////////////////////////////////////////////
// Destructeur d'un Reacteur
/////////////////////////////////////////////////////////////////////////////
Reacteur_impl::~Reacteur_impl() {
	int i;
	if (barres) {
		for (i=0;i<p.nbarres*p.nbarres;++i) barres[i]->Detruire();
		delete [] barres;
	}
	if (canaux) {
		for (i=0;i<2*(p.nbarres*p.nbarres+p.nbarres);++i) canaux[i]->Detruire();
		delete [] canaux;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Duree totale de la simulation
/////////////////////////////////////////////////////////////////////////////
TimeType Reacteur_impl::TempsSimulation() { return time; }

/////////////////////////////////////////////////////////////////////////////
// Dimension du reacteur
/////////////////////////////////////////////////////////////////////////////
IndiceType Reacteur_impl::NombreDeBarres() { return p.nbarres; }

/////////////////////////////////////////////////////////////////////////////
// Initialisation d'un Reacteur: obtention des IOR, etc.
/////////////////////////////////////////////////////////////////////////////
void Reacteur_impl::Initialisation(const ParametresSimulation &par){
	int i;

	// Premierement, verification des parametres generaux de simulation
	// Les autres parametres sont verifies dans les barres ou les caloporteurs
	p = par;
	if (p.Tfus<0) throw MauvaisParametres(p,"Temperature de fusion negative");
	if (p.nbarres < 1) throw MauvaisParametres(p,"Au moins 1 barre necessaire!");

	// Deuxiemement, obtention des IORs des fabriques
	CORBA::Object_var obj;
	desFabriqueDeCanaux fabCanaux;	// sequences de fabriques
	desFabriqueDeBarres fabBarres;
	fabCanaux.length(0);			// initialement, aucune fabrique disponible
	fabCanaux.length(0);
	
	try {	// Obtention du service de nommage
		obj = orb->resolve_initial_references("NameService");
	} catch (CORBA::ORB::InvalidName &) {
		throw MauvaisParametres(p,"Service de nomage inconnu");
	}
	if(CORBA::is_nil(obj.in())) {
		throw MauvaisParametres(p,"'NameService' est une rfrence d'objet nil");
	}
	CosNaming::NamingContext_var nsRef = CosNaming::NamingContext::_narrow(obj);
	if (CORBA::is_nil(nsRef.in())) {
		throw MauvaisParametres(p,"'NameService' n'est pas un NamingContext");
	}

	CosNaming::BindingList_var bl;	// Lister les fabriques de barres et canaux
	CosNaming::BindingIterator_var bi;

	CosNaming::Name_var unNom = new CosNaming::Name();	// Un nom d'objet
	unNom->length(1);

	nsRef->list(1,bl,bi);			// recherche dans les noms disponibles
	if (!CORBA::is_nil(bi)) {
		do {
			unNom[0] = bl[0].binding_name[0];
			if (unNom[0].id[0] == 'c' || unNom[0].id[0] == 'C') {	// fabrique de canaux!
				try {
					obj = nsRef->resolve(unNom);
					if (obj->_non_existent()) {
						nsRef->unbind(unNom);	// Nom invalide, on l'enleve!
					} else {
						fabCanaux.length(fabCanaux.length()+1);	// ajout de la fabrique
						fabCanaux[fabCanaux.length()-1] = FabriqueDeCanaux::_narrow(obj);
					}
				} catch(const CORBA::Exception&) {
					nsRef->unbind(unNom);	// Nom invalide, on l'enleve!

				} catch ( ... ) {
					throw MauvaisParametres(p,"Erreur du service de nommage!");
				}
			}
			else if (unNom[0].id[0] == 'b' || unNom[0].id[0] == 'B') {	// . de barres!
				try {
					obj = nsRef->resolve(unNom);
					if (obj->_non_existent()) {
						nsRef->unbind(unNom);	// Nom invalide, on l'enleve!
					} else {
						fabBarres.length(fabBarres.length()+1);	// ajout de la fabrique
						fabBarres[fabBarres.length()-1] = FabriqueDeBarres::_narrow(obj);
					}
				} catch(const CORBA::Exception&) {
					nsRef->unbind(unNom);	// Nom invalide, on l'enleve!
				} catch ( ... ) {
					throw MauvaisParametres(p,"Erreur du service de nommage!");
				}
			}
		} while (bi->next_n(1,bl));
		bi->destroy();
	}
	cout << "Utilisation de " << fabCanaux.length()
		 << " fabrique(s) de canaux et " << fabBarres.length()
		 << " fabrique(s) de barres." << endl;

	if (fabCanaux.length()==0) {
		throw MauvaisParametres(p,"Aucune fabrique de canaux disponible!");
	}
	if (fabBarres.length()==0) {
		throw MauvaisParametres(p,"Aucune fabrique de barres disponible!");
	}

	// Troisiemement, initialisation generale de la simulation
	time = 0;
	if (barres) delete [] barres;
	if (canaux) delete [] canaux;
	barres = new Barre_var[p.nbarres*p.nbarres];
	canaux = new Canal_var[2*(p.nbarres*p.nbarres+p.nbarres)];

	// Quatriemement, creation et initialisation des objets
	IndiceType fabCourante = 0;	// fabrique utilisee pour l'objet courant
	for (i=0; i<2*(p.nbarres*p.nbarres+p.nbarres); ++i) {	// 2(nb^2+nb) canaux
		canaux[i] = fabCanaux[fabCourante++]->creerObjet();
		canaux[i]->Initialisation(p);
		if (fabCourante == fabCanaux.length()) fabCourante = 0;	// rotation fabriks
	}
	fabCourante = 0;
	for (i=0; i<p.nbarres*p.nbarres; ++i) {					// nb^2 barres
		barres[i] = fabBarres[fabCourante++]->creerObjet();
		barres[i]->Initialisation(p);
		if (fabCourante == fabBarres.length()) fabCourante = 0;	// rotation fabriks
	}

}

/////////////////////////////////////////////////////////////////////////////
// Obtenir un pointeur sur une barre: verification des indices
/////////////////////////////////////////////////////////////////////////////
Barre_var Reacteur_impl::get_barre(IndiceType i, IndiceType j) const {
	if (i>p.nbarres || j>p.nbarres || i<1 ||j<1) {
		Reacteur::IndiceSeq indices;
		indices.length(2);
		indices[0] = i; indices[1] = j;
		throw Reacteur::MauvaisIndice(indices,"Acces a une barre");
	}

	return barres[(i-1)*p.nbarres+(j-1)];	// recupere le bon objet ds le tableau
}

/////////////////////////////////////////////////////////////////////////////
// Obtenir un pointeur sur un canal: verification des indices
/////////////////////////////////////////////////////////////////////////////
Canal_var Reacteur_impl::get_canal(IndiceType i,
									   IndiceType j, IndiceType k) const {
	if ( i > p.nbarres+1 || j>p.nbarres+1 || i<1 || j<1 || k>2 || k<1
			|| (j==p.nbarres+1 && k==2) || (i==p.nbarres+1 && k==1)
			|| (i==p.nbarres+1 && j==p.nbarres+1) )
	{
		Reacteur::IndiceSeq indices;
		indices.length(3);
		indices[0] = i; indices[1] = j; indices[2] = k;
		throw Reacteur::MauvaisIndice(indices,"Acces a un canal");
	}

	if (k==1) return canaux[(i-1)*(p.nbarres+1)+(j-1)];	// pour k=1
	int offset=p.nbarres*p.nbarres+p.nbarres;	// pour k=2
	return canaux[offset+(i-1)*p.nbarres+(j-1)];	// recupere bon objet
}

/////////////////////////////////////////////////////////////////////////////
// Obtenir la temperature d'une barre
/////////////////////////////////////////////////////////////////////////////
TempType Reacteur_impl::temp(IndiceType i, IndiceType j) {
	return get_barre(i,j)->temp_max();		// acces a la methode de barre
}

/////////////////////////////////////////////////////////////////////////////
// Obtenir la temperature d'un canal caloporteur
/////////////////////////////////////////////////////////////////////////////
TempType Reacteur_impl::calop(IndiceType i, IndiceType j, IndiceType k) {
	return get_canal(i,j,k)->temp_moy();	// acces a la methode de canal
}

/////////////////////////////////////////////////////////////////////////////
// Calculer une iteration de 'pas' secondes
/////////////////////////////////////////////////////////////////////////////
void Reacteur_impl::Iteration(TimeType pas) {
	int offset=p.nbarres*p.nbarres+p.nbarres; // Debut des indices calop k=2
	int i,j,k;
	time += pas;

    // Creer un tableau des gradients de temperature actuelles (au temps i)
	TempType *grad_bar = new TempType[p.nbarres*p.nbarres*4];
	Barre::TempArray_slice* grad;

	for (i=0;i<p.nbarres;++i) for (j=0;j<p.nbarres;++j) {	// valeurs actuelles
		grad = get_barre(i+1,j+1)->grad_temp();				// du gradient
		for (k=0;k<4;++k) grad_bar[i*p.nbarres*4+j*4+k] = grad[k];
	}

	// Calculer les nouvelles valeurs de temperature dans les barres
	Reacteur::TempFusionSeq lesFusions(0);
	Barre::TempArray temp_cal;
	for (i=1;i<=p.nbarres;++i) for (j=1;j<=p.nbarres;++j) {
		temp_cal[0] = calop(i,j,1);   temp_cal[1] = calop(i,j,2);
		temp_cal[2] = calop(i,j+1,1); temp_cal[3] = calop(i+1,j,2);
		get_barre(i,j)->Iteration(pas,temp_cal);
		if (get_barre(i,j)->temp_max() >= p.Tfus) {	// En cas de fusion
			lesFusions.length(lesFusions.length()+1);	// la noter.....
			lesFusions[lesFusions.length()-1].temp = get_barre(i,j)->temp_max();
			lesFusions[lesFusions.length()-1].i = i;
			lesFusions[lesFusions.length()-1].j = j;
			lesFusions[lesFusions.length()-1].instant = time;
		}
	}
  
	// Calculer les nouvelles temperatures dans les canaux sur les bords
	for (i=0;i<p.nbarres;++i) {
		// j=1, i=toutes les lignes, k=1 ; gradient barre (i,1) face 1
		get_canal(i+1,1,1)->Iteration(pas,0,grad_bar[i*p.nbarres*4]);
		// j=p.nbarres+1, i=toutes les lignes, k=1 ; gradient barre (i,p.nbarres) face 3
		get_canal(i+1,p.nbarres+1,1)->Iteration(pas,grad_bar[i*p.nbarres*4+(p.nbarres-1)*4+2],0);
		// i=1, j=toutes les colonnes, k=2 ; gradient barre (1,j) face 2
		get_canal(1,i+1,2)->Iteration(pas,0,grad_bar[i*4+1]);  // i=1
		// i=p.nbarres+1, j=toutes les colonnes, k=2 ; gradient barre (p.nbarres,j) face 4
		get_canal(p.nbarres+1,i+1,2)->Iteration(pas,grad_bar[(p.nbarres-1)*p.nbarres*4+i*4+3],0);
	}
	
	for (i=0;i<p.nbarres;++i) for (j=1;j<p.nbarres;++j) { // canaux pour k=1
		// j=2--, i=toutes les lignes, k=1 ; gradient barre (i,j-1) face 3
		//                                         et barre (i,j) face 1
		get_canal(i+1,j+1,1)->Iteration(pas,grad_bar[i*p.nbarres*4+(j-1)*4+2],grad_bar[i*p.nbarres*4+j*4]);
	}

	for (i=1;i<p.nbarres;++i) for (j=0;j<p.nbarres;++j) { // canaux pour k=2
		// j=toutes les colonnes, i=2--, k=2 ; gradient barre (i-1,j) face 4
		//                                           et barre (i,j) face 2
		get_canal(i+1,j+1,2)->Iteration(pas,grad_bar[(i-1)*p.nbarres*4+j*4+3],grad_bar[i*p.nbarres*4+j*4+1]);
	}

	delete [] grad_bar;		// desallouer les valeurs a t-1

	if (lesFusions.length()) {
		throw Reacteur::Fusion(lesFusions);	// En cas de fusion, lever exception
	}
}
